
// first difference approximation

#include <stdio.h>
#include <math.h>

#if 1
double f(double x) { return x*x*x + sqrt(x); }
#define T "double"
const int N=11;
#else
float f(float x) { return x*x*x + sqrtf(x); }
#define T "float"
const int N=7;
#endif

// derivative
double derf(double x) { return 3*x*x + 1.0/(2*sqrt(x)); }

// forward difference
double fd(double x, double dx) { return (f(x+dx)-f(x))/dx; }
// central difference
double fdc(double x, double dx) { return (f(x+dx)-f(x-dx))/(2*dx); }

int main(void) {
    double x = 3;
    double exact = derf(x);
    double dx=1.0;
    printf("# T = %s\n", T);
    printf("# x = %g    der(x) = %.10f\n", x, exact);
    printf("# dx    forward        err         central        err \n");
    for(int i=0; i<N; ++i) {
        dx/=10;
        printf(" %.0e ", dx);
        printf(" %.10f % .3e  ", fd(x,dx), fd(x,dx)-exact);
        printf(" %.10f % .3e\n", fdc(x,dx), fdc(x,dx)-exact);
    }
    printf("\n");
}
